! include for _REAL_ definition, compatible with runmd.F90
#include "../include/dprec.fh"

! module for "middle" scheme for MD / PIMD
!
! Reference:
! Z. Zhang#, X. Liu#, Z. Chen, H. Zheng, K. Yan, and J. Liu*,
! "A unified thermostat scheme for efficient configurational sampling
! for classical/quantum canonical ensembles via molecular dynamics",
! The Journal of Chemical Physics 147, 034109 (2017) 
module md_scheme
  implicit none
  ! the scheme for integration algorithm
  integer :: ischeme     
  integer,parameter :: &
    ENUM_NO_SCHEME       = 0, & ! default AMBER
    ENUM_LFMIDDLE_SCHEME = 1, & ! leap-frog middle: [p-x-T-x]
    ENUM_VVMIDDLE_SCHEME = 2    ! velocity-verlet middle: [p-x-T-x-p]

  ! thermostat options
  integer :: ithermostat ! thermostat method
  integer,parameter :: &
    ENUM_NO_THERM   = 0, & ! no thermostatting
    ENUM_LGV_THERM  = 1, & ! Langevin thermostat
    ENUM_ADS_THERM  = 2, & ! Andersen thermostat
    ENUM_NHC_THERM  = 3    ! Nose-Hoover chain thermostat

  ! thermostat parameter 
  ! = gamma for Langevin dynamics
  ! = nu for Andersen thermostat
  ! = 1/tau for Nose-Hoover chain
  _REAL_ :: therm_par    

  ! see iskip_start and iskip_end in runmd.F90

#include "../include/md.h"

contains
  !-- integration step for the thermostat
  ! @param (in AMBER internal unit)
  !   [inout] v, velocity array (for each DOF)
  !   [in] winv, inverse of mass array (for each atom)
  !   [in] dtx, time step, see runmd.F90
  !   [in] istart, start index of atom, see runmd.F90
  !   [in] iend, end index of atom, see runmd.F90
  ! @external
  !   amrand: random number in uniform dist., in ../lib/random.F90
  !   gauss:  random number in gaussian dist., in ../lib/random.F90
  subroutine thermostat_step(v, winv, dtx, istart, iend, iskip_start, iskip_end)
    use constants, only: KB
    use random
#ifdef LES
    use les_data, only: cnum, temp0les
#endif    
    implicit none

    character(kind=1,len=*),parameter :: routine="thermostat_step"
    _REAL_ :: v(3*nrp), winv(nrp), dtx
    integer :: istart,iend, iskip_start,iskip_end

    _REAL_ :: rand, sigma_i, dt
    integer :: i, j, idof

    _REAL_ :: lgv_c1, lgv_c2, ads_prob, sigma
#ifdef LES
    _REAL_ :: vscalt
#endif    

    dt = dtx / 20.455d0
    select case (ithermostat)
    case (ENUM_LGV_THERM) ! Langevin thermostat
      ! therm_par in ps^-1, dt in ps
      lgv_c1 = exp(-therm_par*dt) 
      lgv_c2 = sqrt(1.0d0 - lgv_c1*lgv_c1)
      sigma = sqrt(kB*temp0)

      do i = 1, iskip_start
        call gauss(0.0d0, 1.0d0, rand)
      end do
      do i = istart, iend
        ! generate random number with normal distribution
        ! gauss is defined in ../lib/random.F90
        ! the unit of sigma is Angstrom per 1/20.455 ps, same as vel
        ! 1 Angstrom per 1/20.455 ps = 1 sqrt(kcal/mol / amu)
        sigma_i = sigma * sqrt(winv(i))
#ifdef LES
        if (temp0les >= 0 .and. cnum(i) > 0) then
          sigma_i = sigma_i * sqrt(temp0les/temp0)
        end if
#endif
        do j = 1, 3
          call gauss(0.0d0, sigma_i, rand) 
          idof = 3*(i-1) + j
          v(idof) = lgv_c1 * v(idof) + lgv_c2 * rand
        end do
      end do
      do i = 1, iskip_end
        call gauss(0.0d0, 1.0d0, rand)
      end do
    case (ENUM_ADS_THERM) ! Andersen thermostat
      ads_prob = 1.0d0 - exp(-therm_par*dt)
      sigma = sqrt(kB*temp0)
      call amrand(rand)
      if (rand < ads_prob) then
        do i = 1, iskip_start
          call gauss(0.0d0, 1.0d0, rand)
        end do

        do i = 1, nrp ! reset all atoms
          sigma_i = sigma * sqrt(winv(i))
          do j = 1, 3
            idof = 3*(i-1) + j
            call gauss(0.0d0, sigma_i, rand) 
            v(idof) = rand
          end do
        end do

        do i = 1, iskip_end
          call gauss(0.0d0, 1.0d0, rand)
        end do
#ifdef LES
        if (temp0les >= 0.0 .and. temp0 .ne. temp0les) then
          vscalt = sqrt(temp0les/temp0)
          do i = 1, nrp
            if (cnum(i) > 0) then
              do j = 1,3
                idof = 3*(i-1)+j
                v(idof) = v(idof) * vscalt
              end do
            end if
          end do
        end if 
#endif      
      end if
    case (ENUM_NHC_THERM) ! Nose-Hoover chain
      write(*,*) '@TODO Nose Hoover chain thermostat'
      stop
    case default
      write(*,*) 'Error in '//routine//': unknown ithermostat'
      stop
    end select
  end subroutine

end module
