#include "../include/dprec.fh"
module qm2_extern_tm_module
! ----------------------------------------------------------------
! Interface for TM based QM and QM/MM MD 
! 
! Orca module interface modified for Turbomole 7.0+ by Holger Kruse 
! Last update Jan'19
!
! Currently supports:
! pure QM
! QM/MM with cutoff for QM-MM electrostatics under periodic
! boundary conditions
!
! ----------------------------------------------------------------

  use qm2_extern_util_module, only: debug_enter_function, debug_exit_function

  implicit none

  private
  public :: get_tm_forces

  character(len=*), parameter :: module_name = "qm2_extern_tm_module"

! there is no easy TM template.
! either run define before or script using cefine string
  type tm_nml_type
   character(len=120) :: cefine
   character(len=80) :: emodule
   character(len=80) :: gmodule
     integer :: ntpr
     integer :: debug
     logical :: dipole
     logical :: docefine

     logical :: dogcp
     character(len=80) :: gcp
     
     ! Deprecated
     integer :: charge
     integer :: spinmult
  end type tm_nml_type


contains

  ! --------------------------------------
  ! Get QM energy and forces from Turbomole
  ! --------------------------------------
  subroutine get_tm_forces( do_grad, nstep, ntpr_default, id, &
       nqmatoms, qmcoords, qmtypes, nclatoms, clcoords, escf, dxyzqm, dxyzcl, &
       charge, spinmult )

    use qm2_extern_util_module, only: print_results, check_installation, write_dipole
    use constants,              only: CODATA08_AU_TO_KCAL, CODATA08_A_TO_BOHRS, ZERO

    logical, intent(in) :: do_grad              ! Return gradient/not
    integer, intent(in) :: nstep                ! MD step number
    integer, intent(in) :: ntpr_default         ! frequency of printing
    character(len=3), intent(in) :: id          ! ID number for PIMD or REMD
    integer, intent(in) :: nqmatoms             ! Number of QM atoms
    _REAL_,  intent(in) :: qmcoords(3,nqmatoms) ! QM coordinates
    integer, intent(in) :: qmtypes(nqmatoms)    ! QM atom types (nuclear charge in au)
    integer, intent(in) :: nclatoms             ! Number of MM atoms
    _REAL_,  intent(in) :: clcoords(4,nclatoms) ! MM coordinates and charges in au
    _REAL_, intent(out) :: escf                 ! SCF energy
    _REAL_, intent(out) :: dxyzqm(3,nqmatoms)   ! SCF QM force
    _REAL_, intent(out) :: dxyzcl(3,nclatoms)   ! SCF MM force
    _REAL_              :: dipxyz(3), dipole    ! Dipole moment
    integer, intent(in) :: charge, spinmult     ! Charge and spin multiplicity

    type(tm_nml_type), save :: tm_nml
    logical, save :: first_call = .true.
    integer :: i
    integer :: printed =-1 ! Used to tell if we have printed this step yet 
                           ! since the same step may be called multiple times
    character(len=512) :: call_buffer
!    character(len=*), parameter  :: program='ridft'
    character(len=*),  parameter :: basename = 'tm_job'
    character(len=*),  parameter :: runext = '.inp'
    character(len=*),  parameter :: datext = '.out'
    character(len=*),  parameter :: dipext = '.dip'
    character(len=*),  parameter :: tplext = '.tpl'
    character(len=14) :: inpfile, datfile, gbwfile, dipfile, crdfile, chgfile, tplfile
    ! Need to prepend subdirectory if doing REMD, PIMD or multi-region QM/MM. 
    !   This is triggered if 'id' is defined (not empty). 
    character(len=25) :: subdir 
    character(len=256), save :: tm_path,gcp_path
    
    datfile = basename//trim(id)//datext
    dipfile = basename//trim(id)//dipext 
    crdfile = 'coord'
    chgfile = 'ptchrg.xyz'
    
    ! Setup on first program call
    if ( first_call ) then
      first_call = .false.
      write (6,'(/,a,/)') '   >>> Running QM/MM calculation with Turbomole <<<'
      write (6,'(/,a,/)') '      >>>        v.1.1 H. Kruse 01.2019    <<<'
      call get_namelist(ntpr_default, tm_nml )
      call check_installation( trim(tm_nml%emodule), id, .true., tm_nml%debug, path=tm_path )
      if(tm_nml%dogcp) then
        write (6,'(/,a,/)') '   >>> applying gCP correction <<<'
      call check_installation('gcp', id, .true., tm_nml%debug, path=gcp_path )
      endif
      call print_namelist( tm_nml )
      if(tm_nml%docefine) then
          write (6,'(80a)') 'Using cefine to generate input!'
          ! cleanup problematic files
          write (6,'(/,a,/)') 'removing: coord control mos alpha beta energy gradient '
          call system('rm -f coord control mos alpha beta energy gradient')
          call write_tm_coord(trim(crdfile),nqmatoms, qmcoords, qmtypes)
          call system(tm_nml%cefine//' > cefine.out 2>&1')
      else
          write (6,'(80a)') 'using provided control file'
       
      endif
      ! write point charge keywords to control file for e-embedding
      if(nclatoms>0)  call mod_control(trim(chgfile))

    end if

     call update_tmfiles(trim(chgfile),trim(crdfile),nqmatoms, qmcoords, qmtypes, nclatoms, clcoords, tm_nml)


    ! Run turbomole modules
    ! energy
    call_buffer=''
    write(call_buffer,'(2a)') &
         trim(call_buffer), trim(tm_path)//'/'//trim(tm_nml%emodule)//'  > '//datfile//' 2>&1'
     if ( tm_nml%debug > 0 ) then
       write (6,'(2a)') 'call_buffer=', trim(call_buffer)
    end if
    call system(trim(call_buffer))
    ! gradient
    call_buffer=''
      write(call_buffer,'(2a)') &
         trim(call_buffer), trim(tm_path)//'/'//trim(tm_nml%gmodule)//'  >> '//datfile//' 2>&1'
    if ( tm_nml%debug > 0 ) then
       write (6,'(2a)') 'call_buffer=', trim(call_buffer)
    end if
     call system(trim(call_buffer))
    !gcp
    if(tm_nml%dogcp) then
     call_buffer=''
       write(call_buffer,'(2a)') &
          trim(call_buffer),trim(gcp_path)//'/gcp '//trim(crdfile)//" -grad -l '"//trim(tm_nml%gcp)//"'  >> "//datfile
     if ( tm_nml%debug > 0 ) then
        write (6,'(2a)') 'call_buffer=', trim(call_buffer)
     end if
      call system(trim(call_buffer))
    endif
    
    

    ! Read tm results
    call read_tm_results(trim(chgfile),nqmatoms, nclatoms, &
    escf, dxyzqm, dxyzcl, dipxyz, dipole, do_grad, tm_nml%dipole, tm_nml%debug )
 
    ! Write dipole moment to dipfile
    if ( tm_nml%ntpr > 0 .and. mod(nstep, tm_nml%ntpr) == 0 ) then
      if ( printed /= nstep .and. tm_nml%dipole ) then
        call write_dipole(trim(dipfile), dipxyz, dipole, tm_nml%debug)
        printed = nstep
      end if
    end if

    ! Backup old input and datfiles (if in subdirectory, copied out earlier)
    call system('cp coord old.coord')
    call system('cp '//trim(datfile)//' '//'old.'//datfile)
     if (nclatoms > 0) then
        call system('cp '//trim(chgfile)//' '//'old.'//chgfile)
     end if 

    if ( do_grad ) then
      ! Convert Hartree/Bohr -> kcal/(mol*A)
      dxyzqm(:,:) = dxyzqm(:,:) * CODATA08_AU_TO_KCAL * CODATA08_A_TO_BOHRS
      if ( nclatoms > 0 ) then
         dxyzcl(:,:) = dxyzcl(:,:) * CODATA08_AU_TO_KCAL * CODATA08_A_TO_BOHRS
      end if
    else
      dxyzqm = ZERO
      if ( nclatoms > 0 ) dxyzcl = ZERO
    end if

    escf = escf * CODATA08_AU_TO_KCAL

    call print_results( 'qm2_extern_tm_module', escf, nqmatoms, dxyzqm,&
      tm_nml%debug, nclatoms, dxyzcl )

  end subroutine get_tm_forces

  ! -----------------------------------------------
  ! Read Turbomole namelist values from file mdin,
  ! use default values if none are present.
  ! -----------------------------------------------
  subroutine get_namelist( ntpr_default, tm_nml )
!~ 
    use UtilitiesModule, only: Upcase
    implicit none
!~ 
    integer, intent(in) :: ntpr_default
    type(tm_nml_type), intent(out) :: tm_nml
    character(len=120) :: cefine,emodule,gmodule,gcp
    logical docefine,dogcp
    integer :: ntpr, debug, dipole
    namelist /tm/ cefine,ntpr,debug,emodule,gmodule,docefine,dogcp,gcp
    integer :: ierr
!~ 
!~     ! Default values
       cefine = 'define < define.inp'
       emodule= 'ridft'
       gmodule= 'rdgrad'
       ntpr         = ntpr_default
       debug        = 0
       docefine =.true.
       dogcp =.false.
       gcp = ''
!~ 
!~     ! Read namelist
    rewind 5
    read(5,nml=tm,iostat=ierr)
!~ 
    if ( ierr > 0 ) then
       call sander_bomb('get_namelist (qm2_extern_tm_module)', &
            '&tm namelist read error', &
            'Please check your input.')
    else if ( ierr < 0 ) then
       write(6,'(a,/,a)') '&tm namelist read encountered end of file', &
            'Please check your input if the calculation encounters a problem'
    end if

    tm_nml%emodule       = emodule
    tm_nml%gmodule       = gmodule
    tm_nml%cefine = cefine
    tm_nml%gcp = gcp
    tm_nml%ntpr        = ntpr
    tm_nml%debug       = debug
    tm_nml%docefine = docefine
    tm_nml%dogcp = dogcp

     end subroutine get_namelist

  ! --------------------------------
  ! Print Turbomole namelist settings
  ! --------------------------------
  subroutine print_namelist( tm_nml )

    implicit none
    type(tm_nml_type), intent(in) :: tm_nml


    write(6, '(/,a)')      '| &tm'
    write(6, '(2a)')       '|   define (string)         = ', tm_nml%cefine
    write(6, '(2a)')       '|   gcp    (string)         = ', tm_nml%gcp
    write(6, '(2a)')       '|   energy module (string)  = ', tm_nml%emodule
    write(6, '(2a)')       '|   gradient module (string)= ', tm_nml%gmodule
    write(6, '(a,i0)')     '|   ntpr         = ', tm_nml%ntpr
    write(6, '(a,i2)')     '|   debug        = ', tm_nml%debug
    write(6, '(a,l)')      '|   docefine (logical) = ', tm_nml%docefine
    write(6, '(a,l)')      '|   dogcp (logical) = ', tm_nml%dogcp
    write(6,'(a)')         '| /'

  end subroutine print_namelist

  subroutine update_tmfiles(chgfile,crdfile,nqmatoms, qmcoords, qmtypes, nclatoms, clcoords, tm_nml)
    use ElementOrbitalIndex,    only: elementSymbol
    use qm2_extern_util_module, only: write_chgfile
    use constants,              only: CODATA08_AU_TO_KCAL, CODATA08_A_TO_BOHRS
    implicit none

    character (len=*) , intent(in) :: crdfile
    character (len=*) , intent(in) :: chgfile
    character (len=120)            :: string
    integer           , intent(in) :: nqmatoms
    _REAL_            , intent(in) :: qmcoords(:,:)
    integer           , intent(in) :: qmtypes(:)
    integer           , intent(in) :: nclatoms
    _REAL_            , intent(in) :: clcoords(:,:)
    type(tm_nml_type), intent(in) :: tm_nml
      _REAL_ f
    integer            :: i, ios
    integer, parameter :: iurun = 10
    logical, save :: first_call = .true.


call debug_enter_function( 'write_inpfile', module_name, tm_nml%debug )

! turbomole wants point charge aufpunkt in bohr!
  f=CODATA08_A_TO_BOHRS
  call write_tm_coord(crdfile,nqmatoms, qmcoords, qmtypes)

 
  if(nclatoms>0) then
! write pointcharges into file
  open(23,file=chgfile)
  write(23,'(a)') '$point_charges'
     do i = 1, nclatoms
         write(23,'(4f21.16)') clcoords(1:3,i)*f,clcoords(4,i)
      enddo
  write(23,'(a)') '$end'
  close(23)
  endif

end subroutine update_tmfiles


subroutine write_tm_coord(crdfile,nqmatoms, qmcoords, qmtypes)
    use ElementOrbitalIndex,    only: elementSymbol
    use constants,              only: CODATA08_AU_TO_KCAL, CODATA08_A_TO_BOHRS, ZERO
    implicit none

    _REAL_             :: f
    integer            :: i, ios
    integer           , intent(in) :: nqmatoms
    _REAL_            , intent(in) :: qmcoords(:,:)
    integer           , intent(in) :: qmtypes(:)
    character (len=*) , intent(in) :: crdfile

   f=CODATA08_A_TO_BOHRS
  ! write new coord file.
  open(22,file=crdfile)
  write(22,'(a)') '$coord'
  do i=1,nqmatoms
  write(22,'(3F18.12,2x,a2)') qmcoords(1:3,i)*f, elementSymbol(qmtypes(i))
  enddo
  write(22,'(a)') '$end'
  close(22)

return
end subroutine


  subroutine mod_control(chgfile)
  implicit none
  character(80) string
  character (len=*) , intent(in) :: chgfile

! set point charge options in TM, as well as request ptch deriv.

  call system('kdg end')
  call system('kdg drvopt')
  call system('kdg point_charges')
  call system('kdg point_charge_gradients') 
  open(33,file='control',status='old',position='append')
  write(33,'(a)') '$point_charges file='//chgfile//' nocheck  '
  write(33,'(a)') '$point_charge_gradients  file='//chgfile//'.pcgrad '
  write(33,'(a)') '$drvopt'
  write(33,'(a)') '  cartesian  on'
  write(33,'(a)') '  basis      off'
  write(33,'(a)') '  global     off'
  write(33,'(a)') '  hessian    on'
  write(33,'(a)') '  dipole     on'
  write(33,'(a)') ' nuclear polarizability'
  write(33,'(a)') ' point charges'
  write(33,'(a)') '$end'
  close(33)
     

  end subroutine mod_control

  ! ---------------------
  ! Read turbomole results
  ! ---------------------
  subroutine read_tm_results(chgfile,nqmatoms, nclatoms, &
    escf, dxyzqm, dxyzcl, dipxyz, dipole, do_grad, do_dipole, debug )

    use constants, only:  CODATA08_AU_TO_DEBYE

    implicit none

     character(len=*), intent(in) :: chgfile
    ! Used to get the .engrad and .pcgrad files
!~     character(len=*), intent(in) :: basename
    integer, intent(in) :: nqmatoms, nclatoms 
    _REAL_, intent(out) :: escf, dxyzqm(3,nqmatoms), dxyzcl(3,nclatoms)
    logical, intent(in) :: do_grad, do_dipole
    integer, intent(in) :: debug

    integer :: ios, i,idx
    integer, parameter :: iunit = 351
    _REAL_, intent(out) :: dipxyz(3), dipole ! return in Debye
    character(len=256) :: read_buffer

    call debug_enter_function( 'read_results', module_name, debug )

    if ( do_grad ) then
      ! Energy and gradient
      open(iunit, file='gradient', iostat=ios)
      if ( ios /= 0 ) then
        call sander_bomb('read_results (qm2_extern_tm_module)', &
          'Error opening Turbomole output file <gradient> for reading', &
          'Will quit now')
      end if
      do
        read (iunit, '(a)',end=666, iostat = ios) read_buffer
        ! End of file; data not found
        if (ios < 0) then
          call sander_bomb('read_results (qm2_extern_tm_module)', &
            'Error reading Turbomole output from file <gradient>', &
            '(Current total energy or gradient not found or unsupported units.)')
        end if


         ! read energy+gradient from gradient file. Different formats make this a pain. 
         ! We identify the 2nd '=' to find the start of the buffer. buffer seems always 20 char long.
         if (index(read_buffer,"cycle")/=0) then
             idx=0
             do i=1,len(read_buffer)
               if(read_buffer(i:i)=='=') idx=idx+1
               if(idx==2) exit
             enddo
!           print*,idx,i,i+1
           idx=i+1
           read(read_buffer(idx:idx+20),*) escf
           if ( debug > 0 ) then
              write (6,*) 'energy, read_buffer indices =', escf, idx,idx+20
           end if
!           read(read_buffer(33:53),*) escf
           do i=1,nqmatoms
               read (iunit,'(A)') read_buffer ! dummy read for coordinates
           enddo
           do i=1,nqmatoms
               read (iunit,*) dxyzqm(1,i), dxyzqm(2,i), dxyzqm(3,i)
           enddo
         endif


      end do
666 continue
      close(iunit)

      ! MM gradients (none if do_grad)
! MODIFY
! for pointcharges
      if ( nclatoms > 0 ) then
        open(iunit, file=chgfile//'.pcgrad', iostat=ios)
        do
          read (iunit, '(a)', iostat = ios) read_buffer !First line is $point_charge_gradients
          ! Read MM gradient data
          do i = 1, nclatoms
            read (iunit, *) dxyzcl(:,i)
!~              write(*,*) dxyzcl(:,i)
          end do
          exit
        end do
        close(iunit)
      end if ! nclatoms if
    end if ! do_grad if



    dipole=0d0
    dipxyz=0d0
    if ( do_dipole .or. (.not. do_grad) ) then
    stop 'NOT IMPLEMENTED SRY!'
!~       ! Dipole moment
!~       open(iunit, file=datfile, iostat=ios)
!~       if ( ios /= 0 ) then
!~         call sander_bomb('read_results (qm2_extern_orc_module)', &
!~           'Error opening Orca output file '//datfile//' for reading', &
!~           'Will quit now')
!~       end if
!~       do
!~         read (iunit, '(a)', iostat = ios) read_buffer
!~         ! End of file; data not found
!~         if (ios < 0) then
!~           call sander_bomb('read_results (qm2_extern_orc_module)', &
!~             'Error reading Orca output from file '//datfile, &
!~             '(Dipole moment not found.)')
!~         end if
!~         if ( .not. do_grad .and. index(read_buffer, 'FINAL SINGLE POINT')>0 ) then
!~           read(read_buffer(29:),'(f22.12)') escf
!~         end if
!~         if (index(read_buffer,'Total Dipole Moment') > 0) then
!~           read(read_buffer(29:),*) dipxyz(:)
!~           read (iunit, '(a)') read_buffer ! Skip
!~           read (iunit, '(a)') read_buffer
!~           read(read_buffer(29:),*) dipole
!~           ! convert au to Debye
!~           dipxyz(:) = dipxyz(:) * CODATA08_AU_TO_DEBYE
!~           dipole = dipole * CODATA08_AU_TO_DEBYE
!~           exit
!~         end if
!~       end do
!~       close(iunit)
    end if

    call debug_exit_function( 'read_results', module_name, debug )

  end subroutine read_tm_results

end module qm2_extern_tm_module
