#include "../include/dprec.fh"
module qm2_extern_xtb_module
! ----------------------------------------------------------------
! Interface for GFN-xTB based QM and QM/MM MD 
! H.Kruse Oct.2017
! Currently supports:
! pure QM
! QM/MM with cutoff for QM-MM electrostatics under periodic
! boundary conditions
! no dipoles
! ----------------------------------------------------------------

  use qm2_extern_util_module, only: debug_enter_function, debug_exit_function

  implicit none

  private
  public :: get_xtb_forces

  character(len=*), parameter :: module_name = "qm2_extern_xtb_module"

  type xtb_nml_type
   character(len=120) :: flags
   character(len=120) :: prog
     integer :: ntpr
     integer :: debug
     integer :: dipole
     integer :: do_pcel
  end type xtb_nml_type


contains

  ! --------------------------------------
  ! Get QM energy and forces from xtb (similar to turbomole!)
  ! --------------------------------------
  subroutine get_xtb_forces( do_grad, nstep, ntpr_default, id, &
       nqmatoms, qmcoords, qmtypes, nclatoms, clcoords, escf, dxyzqm, dxyzcl, &
       charge, spinmult )

    use qm2_extern_util_module, only: print_results, check_installation, write_dipole
    use constants,              only: CODATA08_AU_TO_KCAL, CODATA08_A_TO_BOHRS, ZERO

    logical, intent(in) :: do_grad              ! Return gradient/not
    integer, intent(in) :: nstep                ! MD step number
    integer, intent(in) :: ntpr_default         ! frequency of printing
    character(len=3), intent(in) :: id          ! ID number for PIMD or REMD
    integer, intent(in) :: nqmatoms             ! Number of QM atoms
    _REAL_,  intent(in) :: qmcoords(3,nqmatoms) ! QM coordinates
    integer, intent(in) :: qmtypes(nqmatoms)    ! QM atom types (nuclear charge in au)
    integer, intent(in) :: nclatoms             ! Number of MM atoms
    _REAL_,  intent(in) :: clcoords(4,nclatoms) ! MM coordinates and charges in au
    _REAL_, intent(out) :: escf                 ! SCF energy
    _REAL_, intent(out) :: dxyzqm(3,nqmatoms)   ! SCF QM force
    _REAL_, intent(out) :: dxyzcl(3,nclatoms)   ! SCF MM force
    _REAL_              :: dipxyz(3), dipole    ! Dipole moment
    integer, intent(in) :: charge, spinmult     ! Charge and spin multiplicity

    type(xtb_nml_type), save :: xtb_nml
    logical, save :: first_call = .true.
    integer :: i
    integer :: printed =-1 ! Used to tell if we have printed this step yet 
                           ! since the same step may be called multiple times
    character(len=512) :: call_buffer
    character(len=*),  parameter :: basename = 'xtb_job'
    character(len=*),  parameter :: runext = '.inp'
    character(len=*),  parameter :: datext = '.out'
    character(len=*),  parameter :: dipext = '.dip'
    character(len=*),  parameter :: tplext = '.tpl'
    character(len=14) :: inpfile, datfile, gbwfile, dipfile, crdfile, chgfile, tplfile
    ! Need to prepend subdirectory if doing REMD, PIMD or multi-region QM/MM. 
    !   This is triggered if 'id' is defined (not empty). 
    character(len=25) :: subdir
    character(len=3)  :: strchrg 
    character(len=256), save :: xtb_path,cla
    
    ! assemble input - / output data filenames
    datfile = basename//trim(id)//datext
    dipfile = basename//trim(id)//dipext 
    crdfile = 'coord.xyz'
    chgfile = 'pcharge'
    
    ! Setup on first program call
    if ( first_call ) then
      first_call = .false.
      write (6,'(/,a,/)') '   >>> Running QM/MM calculation with xTB <<<'
      write (6,'(a)') '      >>>        v.1.1 H. Kruse 01.2019    <<<'
      call get_namelist(ntpr_default, xtb_nml )
      call check_installation( trim(xtb_nml%prog), id, .true., xtb_nml%debug, path=xtb_path )
      call print_namelist( xtb_nml )
      write (6,'(/,a,/)') 'removing old files: coord.xyz pcharge pcgrad energy gradient xtbrestart'
      call system('rm -f pcgrad coord.xyz pcharge energy gradient pcgrad xtbrestart')
      call write_xyz_coord(trim(crdfile),nqmatoms, qmcoords, qmtypes)
      write(strchrg,'(I3)') charge
      cla=' '//trim(crdfile)//' '//trim(xtb_nml%flags)//' --chrg '//trim(adjustl(strchrg))
      if(do_grad) then
        cla=trim(cla)//' --grad'
      endif
      write (6,'(3x,a)') 'xtb cli:',trim(cla)
    end if

     call update_xtbfiles(trim(chgfile),trim(crdfile),nqmatoms, qmcoords, qmtypes, nclatoms, clcoords, xtb_nml)


    ! Run xtb
    ! energy
    call_buffer=''
    write(call_buffer,'(2a)') &
       trim(call_buffer), trim(xtb_path)//'/'//trim(xtb_nml%prog)//trim(cla)//'  > '//datfile//' 2>&1'
     if ( xtb_nml%debug > 0 ) then
       write (6,'(2a)') 'call_buffer=', trim(call_buffer)
    end if
    call system(trim(call_buffer))

    ! Read xtb results
    call read_xtb_results(trim(chgfile),nqmatoms, nclatoms, &
    escf, dxyzqm, dxyzcl, dipxyz, dipole, do_grad, xtb_nml%dipole, xtb_nml%debug )
 
    ! Backup old input and datfiles (if in subdirectory, copied out earlier)
    call system('cp coord.xyz old.coord.xyz')
    call system('cp '//trim(datfile)//' '//'old.'//datfile)
     if (nclatoms > 0) then
        call system('cp '//trim(chgfile)//' '//'old.'//chgfile)
     end if 

    if ( do_grad ) then
      ! Convert Hartree/Bohr -> kcal/(mol*A)
      dxyzqm(:,:) = dxyzqm(:,:) * CODATA08_AU_TO_KCAL * CODATA08_A_TO_BOHRS
      if ( nclatoms > 0 ) then
         dxyzcl(:,:) = dxyzcl(:,:) * CODATA08_AU_TO_KCAL * CODATA08_A_TO_BOHRS
      end if
    else
      dxyzqm = ZERO
      if ( nclatoms > 0 ) dxyzcl = ZERO
    end if

    escf = escf * CODATA08_AU_TO_KCAL

    call print_results( 'qm2_extern_xtb_module', escf, nqmatoms, dxyzqm,&
      xtb_nml%debug, nclatoms, dxyzcl )

  end subroutine get_xtb_forces

  ! -----------------------------------------------
  ! Read xTB namelist values from file mdin,
  ! use default values if none are present.
  ! -----------------------------------------------
  subroutine get_namelist( ntpr_default, xtb_nml )
    use UtilitiesModule, only: Upcase
    implicit none
    integer, intent(in) :: ntpr_default
    type(xtb_nml_type), intent(out) :: xtb_nml
    character(len=120) :: flags,prog
    integer :: ntpr, debug, dipole, do_pcel
    namelist /xtb/ ntpr,debug,prog,flags,dipole,do_pcel
    integer :: ierr


       prog = 'xtb'
       flags= '--gfn 2 '
       ntpr         = ntpr_default
       debug        = 0
       dipole       = 0
       do_pcel      = 0


    rewind 5
    read(5,nml=xtb,iostat=ierr)

    if ( ierr > 0 ) then
       call sander_bomb('get_namelist (qm2_extern_xtb_module)', &
            '&xtb namelist read error', &
            'Please check your input.')
    else if ( ierr < 0 ) then
       write(6,'(a,/,a)') '&xtb namelist read encountered end of file', &
            'Please check your input if the calculation encounters a problem'
    end if

    xtb_nml%prog       = prog
    xtb_nml%flags       = flags
    xtb_nml%ntpr        = ntpr
    xtb_nml%debug       = debug
    xtb_nml%do_pcel     = do_pcel

    if( xtb_nml%do_pcel > 0) write(6,'(a)') 'Using element symbols for xTB point-charges'
    if ( dipole == 0 ) then
        xtb_nml%dipole = 0
     else if ( dipole == 1 ) then
        xtb_nml%dipole = 1
     else
        call sander_bomb('get_namelist (qm2_extern_orc_module)', &
             '&xtb dipole value not allowed', &
             'Please check your input. dipole can only be 0 or 1.')
     end if
     end subroutine get_namelist

  ! --------------------------------
  ! Print xTB namelist settings
  ! --------------------------------
  subroutine print_namelist( xtb_nml )

    implicit none
    type(xtb_nml_type), intent(in) :: xtb_nml

    write(6, '(/,a)')      '| &xtb'
    write(6, '(2a)')       '|   prog        = ', xtb_nml%prog
    write(6, '(2a)')       '|   flags       = ', xtb_nml%flags
    write(6, '(a,i2)')     '|   debug       = ', xtb_nml%debug
!    write(6, '(a,i2)')     '|   do_pcel     = ', xtb_nml%do_pcel
!    write(6, '(a,i2)')     '|   dipole      = ', xtb_nml%dipole
    write(6,'(a)')         '| /'
  end subroutine print_namelist

  subroutine update_xtbfiles(chgfile,crdfile,nqmatoms, qmcoords, qmtypes, nclatoms, clcoords, xtb_nml)
    use ElementOrbitalIndex,    only: elementSymbol
    use qm2_extern_util_module, only: write_chgfile
    use constants,              only: CODATA08_AU_TO_KCAL, CODATA08_A_TO_BOHRS
    implicit none

    character (len=*) , intent(in) :: crdfile
    character (len=*) , intent(in) :: chgfile
    character (len=120)            :: string
    integer           , intent(in) :: nqmatoms
    _REAL_            , intent(in) :: qmcoords(:,:)
    integer           , intent(in) :: qmtypes(:)
    integer           , intent(in) :: nclatoms
    _REAL_            , intent(in) :: clcoords(:,:)
    type(xtb_nml_type), intent(in) :: xtb_nml
    _REAL_             :: f
    integer            :: i, ios
    integer, parameter :: iurun = 10
    logical, save :: first_call = .true.

  ! xtb also wants point charge aufpunkt in bohr!
  f=CODATA08_A_TO_BOHRS
  call debug_enter_function( 'write_inpfile', module_name, xtb_nml%debug )
  call write_xyz_coord(crdfile,nqmatoms, qmcoords, qmtypes)
   ! write pointcharges into file
  if(nclatoms>0) then
  open(23,file=chgfile)
  write(23,'(I6)') nclatoms
    if(xtb_nml%do_pcel>0) then
     do i = 1, nclatoms
         write(23,'(4(f22.16,x),2x,i2)') clcoords(4,i), clcoords(1:3,i)*f,1
      enddo
    else
     do i = 1, nclatoms
         write(23,'(4(f22.16,x))') clcoords(4,i), clcoords(1:3,i)*f
      enddo
    endif
  close(23)
  endif
end subroutine update_xtbfiles


subroutine write_xyz_coord(crdfile,nqmatoms, qmcoords, qmtypes)
    use ElementOrbitalIndex,    only: elementSymbol
    use constants,              only: CODATA08_AU_TO_KCAL, CODATA08_A_TO_BOHRS, ZERO
    implicit none

    _REAL_             :: f
    integer            :: i, ios
    integer           , intent(in) :: nqmatoms
    _REAL_            , intent(in) :: qmcoords(:,:)
    integer           , intent(in) :: qmtypes(:)
    character (len=*) , intent(in) :: crdfile

  ! f=CODATA08_A_TO_BOHRS
  open(22,file=crdfile)
  write(22,'(I6)') nqmatoms
  write(22,'(a)') 'sander qm-region xyz file'
  do i=1,nqmatoms
    write(22,'(a2,2x,3F18.12)')  elementSymbol(qmtypes(i)),qmcoords(1:3,i)
  enddo
  close(22)

return
end subroutine


  ! ---------------------
  ! Read xTB results (almost same as Turbomole!)
  ! writes same energy/gradient file
  ! ---------------------
  subroutine read_xtb_results(chgfile,nqmatoms, nclatoms, &
    escf, dxyzqm, dxyzcl, dipxyz, dipole, do_grad, do_dipole, debug )

    use constants, only:  CODATA08_AU_TO_DEBYE

    implicit none

     character(len=*), intent(in) :: chgfile
    ! Used to get the .engrad and .pcgrad files
!~     character(len=*), intent(in) :: basename
    integer, intent(in) :: nqmatoms, nclatoms 
    _REAL_, intent(out) :: escf, dxyzqm(3,nqmatoms), dxyzcl(3,nclatoms)
    logical, intent(in) :: do_grad
    integer, intent(in) :: debug, do_dipole

    integer :: ios, i,idx
    integer, parameter :: iunit = 351
    _REAL_, intent(out) :: dipxyz(3), dipole ! return in Debye
    character(len=256) :: read_buffer

    call debug_enter_function( 'read_results', module_name, debug )

    if ( do_grad ) then
      ! Energy and gradient
      open(iunit, file='gradient', iostat=ios)
      if ( ios /= 0 ) then
        call sander_bomb('read_results (qm2_extern_xtb_module)', &
          'Error opening Turbomole output file <gradient> for reading', &
          'Will quit now')
      end if
      do
        read (iunit, '(a)',end=666, iostat = ios) read_buffer
        ! End of file; data not found
        if (ios < 0) then
          call sander_bomb('read_results (qm2_extern_xtb_module)', &
            'Error reading Turbomole output from file <gradient>', &
            '(Current total energy or gradient not found or unsupported units.)')
        end if


         ! read energy+gradient from gradient file. Different formats make this a pain. 
         ! We identify the 2nd '=' to find the start of the buffer. buffer seems always 20 char long.
         if (index(read_buffer,"cycle")/=0) then
             idx=0
             do i=1,len(read_buffer)
               if(read_buffer(i:i)=='=') idx=idx+1
               if(idx==2) exit
             enddo
!           print*,idx,i,i+1
           idx=i+1
           read(read_buffer(idx:idx+20),*) escf
           if ( debug > 0 ) then
              write (6,*) 'energy, read_buffer indices =', escf, idx,idx+20
           end if
!           read(read_buffer(33:53),*) escf
           do i=1,nqmatoms
               read (iunit,'(A)') read_buffer ! dummy read for coordinates
           enddo
           do i=1,nqmatoms
               read (iunit,*) dxyzqm(1,i), dxyzqm(2,i), dxyzqm(3,i)
           enddo
         endif


      end do
666 continue
      close(iunit)

    !  pointcharge gradient
      if ( nclatoms > 0 ) then
        open(iunit, file='pcgrad', iostat=ios)
          do i = 1, nclatoms
            read (iunit, *) dxyzcl(:,i)
!~              write(*,*) dxyzcl(:,i)
          end do
        close(iunit)
      end if ! nclatoms if

    end if ! do_grad if



    dipole=0d0
    dipxyz=0d0
    if ( do_dipole>1 .or. (.not. do_grad) ) then
    stop 'NOT IMPLEMENTED SRY!'
!~       ! Dipole moment
!~       open(iunit, file=datfile, iostat=ios)
!~       if ( ios /= 0 ) then
!~         call sander_bomb('read_results (qm2_extern_orc_module)', &
!~           'Error opening Orca output file '//datfile//' for reading', &
!~           'Will quit now')
!~       end if
!~       do
!~         read (iunit, '(a)', iostat = ios) read_buffer
!~         ! End of file; data not found
!~         if (ios < 0) then
!~           call sander_bomb('read_results (qm2_extern_orc_module)', &
!~             'Error reading Orca output from file '//datfile, &
!~             '(Dipole moment not found.)')
!~         end if
!~         if ( .not. do_grad .and. index(read_buffer, 'FINAL SINGLE POINT')>0 ) then
!~           read(read_buffer(29:),'(f22.12)') escf
!~         end if
!~         if (index(read_buffer,'Total Dipole Moment') > 0) then
!~           read(read_buffer(29:),*) dipxyz(:)
!~           read (iunit, '(a)') read_buffer ! Skip
!~           read (iunit, '(a)') read_buffer
!~           read(read_buffer(29:),*) dipole
!~           ! convert au to Debye
!~           dipxyz(:) = dipxyz(:) * CODATA08_AU_TO_DEBYE
!~           dipole = dipole * CODATA08_AU_TO_DEBYE
!~           exit
!~         end if
!~       end do
!~       close(iunit)
    end if

    call debug_exit_function( 'read_results', module_name, debug )

  end subroutine read_xtb_results

end module qm2_extern_xtb_module
